#include "sara.h"


void processEvents() {
	SDL_Event event;

	while (SDL_PollEvent(&event)) {
		if (event.type == SDL_KEYDOWN) {
			if (event.key.keysym.sym == SDLK_ESCAPE)
				exit(0);
		} else if (event.type == SDL_QUIT)
			exit(0);
	}
}


void initGroups(int groups) {
	info.group = malloc(sizeof(QUAD_GROUP) * groups);
	info.groups = groups;
	
	return;
}


void initGroup(int group, int members) {
	int i;

	info.group[group].quads = members;
	info.group[group].quad = malloc(sizeof(QUAD_INFO) * members);
	info.group[group].texture = renderTextureNew(512, 512, 8, 8);
	info.group[group].qc = renderCacheAlloc(info.group[group].quads, info.group[group].texture);
	for (i = 0; i < members; i++)
		textureNoiseProfileSetup(&info.group[group].quad[i].noise);

	return;
}


float rotateVelGenerate(float max) {
	int i;

	i = rand();
	i &= 0xFFF;
	return max * i / 0xFFF;
}


float coordinateGenerate(float center, float radius) {
	int i;

	i = rand();
	i &= 0xFFF;
	return (center + (float) i / 0xFFF * radius * 2) - radius;
}


void quadsSpawn(float side) {
	int i, j;

	for (i = 0; i < info.groups; i++)
		for (j = 0; j < info.group[i].quads; j++) {
			renderCacheSetOne(info.group[i].qc, j, 
				coordinateGenerate(TILE_CENTER_X, TILE_RADIUS),
				coordinateGenerate(TILE_CENTER_Y, TILE_RADIUS),
				coordinateGenerate(TILE_CENTER_Z, TILE_RADIUS), side, j);
			renderCacheRotateOne(info.group[i].qc, j, rotateVelGenerate(M_PI * 2),
				rotateVelGenerate(M_PI * 2), rotateVelGenerate(M_PI * 2));
		}
	
	return;
}


void targetCalcCoords() {
	int i, j;
	float space_w, space_h, side, skip, x, y;

	space_w = 2.0f / info.w;
	space_h = (float) (2.0f * video.h_mult) / info.h;
	side = (space_w > space_h) ? space_h : space_w;
	skip = side * 2;

	x = (WINDOW_W - skip * info.w) / 2 - (WINDOW_W / 2);
	x += side;
	y = (WINDOW_H - skip * info.h) / 2 - (WINDOW_H / 2);
	y *= -1;
	y -= side;
	side -= QUAD_PADDING;

	for (i = 0; i < info.groups; i++) {
		for (j = 0; j < info.group[i].quads; j++) {
			info.group[i].quad[j].target_x = x + info.group[i].quad[j].xi * skip;
			info.group[i].quad[j].target_y = y - info.group[i].quad[j].yi * skip;
			info.group[i].quad[j].target_z = QUAD_Z_TARGET;
			
			info.group[i].quad[j].rotvel_x = rotateVelGenerate(ROTATE_VEL_X_MAX);
			info.group[i].quad[j].rotvel_y = rotateVelGenerate(ROTATE_VEL_Y_MAX);
			info.group[i].quad[j].rotvel_z = rotateVelGenerate(ROTATE_VEL_Z_MAX);
		}
	}

	quadsSpawn(side);
	info.follow.quad = rand() % info.assemble.quads;
	info.follow.qc = renderCacheAlloc(1, info.group[info.assemble.quad[info.follow.quad].group].texture);
	info.fade_start = 0;
	info.mode = MODE_DUMMY;
	info.newmode = MODE_FOLLOW;
	info.fade_start = SDL_GetTicks() - FADE_TIME;

	return;
}


void targetPatternLoad(const char *path) {
	FILE *fp;
	unsigned int *pattern, w, h, tiles, *pattern_index, groups;
	int i, j, k, t, len, index;

	if ((fp = fopen(path, "r")) == NULL) {
		fprintf(stderr, "Kunde inte öppna mig själv. Slutmönstret kan därför inte läsas in. Systemfel. Etc.\n");
		exit(-1);
	}

	fseek(fp, -4, SEEK_END);
	fread(&len, 4, 1, fp);
	fseek(fp, -1*len, SEEK_END);
	fread(&w, 4, 1, fp);
	len -= 8;
	pattern = malloc(len);
	fread(pattern, len, 1, fp);
	fclose(fp);

	for (i = tiles = 0; i < (len>>2); i++)
		if (pattern[i] != 0)
			tiles++;
	pattern_index = malloc(sizeof(unsigned int) * tiles);
	for (i = tiles = 0; i < (len>>2); i++)
		if (pattern[i] != 0) {
			pattern_index[tiles] = pattern[i];
			tiles++;
		}
	
	/* Sortera för att göra det lättare att identifiera antalet grupper */
	for (i = 1; i < tiles; i++)
		for (j = i; j > 0 && pattern_index[j] < pattern_index[j-1]; j--) {
			t = pattern_index[j];
			pattern_index[j] = pattern_index[j - 1];
			pattern_index[j - 1] = t;
		}
	
	/* Räkna mängden grupper */
	for (i = groups = t = 0; i < tiles; i++)
		if (t != pattern_index[i]) {
			t = pattern_index[i];
			groups++;
		}


	/* Förbered för grupperna */
	initGroups(groups);
	for (i = t = groups = 0; i < tiles; i++)
		if (t != pattern_index[i]) {
			t = pattern_index[i];
			if (t != 0) {
				info.group[groups].type = pattern_index[i];
				groups++;
			}
		}
	

	/* Initiera grupperna */
	for (i = 0; i < groups; i++) {
		for (j = t = 0; j < tiles; j++)
			if (info.group[i].type == pattern_index[j]) 
				t++;
		initGroup(i, t);
	}

	/* Fyll i lite information för varje medlem */
	for (i = h = index = 0; i < groups; i++)
		for (j = t = 0; j < (len>>2); j++)
			if (info.group[i].type == pattern[j]) {
				info.group[i].quad[t].xi = j % w;
				info.group[i].quad[t].yi = j / w;
				info.group[i].quad[t].index = index;
				if (j / w > h)
					h = j / w;
				t++;
				index++;
			}


	/* Fyll i strukturer som behövs när vi slumpar fram vilken platta vi ska flytta till slutpositionen */
	info.assemble.quad = malloc(sizeof(QUAD_POINTER) * tiles);
	info.assemble.quads = tiles;
	info.assemble.current.group = -1;
	info.assemble.current.quad = -1;
	info.assemble.current.assembled = 1;

	
	for (i = k = 0; i < groups; i++)
		for (j = 0; j < info.group[i].quads; j++, k++) {
			index = info.group[i].quad[j].index;
			info.assemble.quad[index].group = i;
			info.assemble.quad[index].quad = j;
			info.assemble.quad[index].assembled = 0;
		}

	/* Sätt de sista värderna */
	free(pattern_index);
	info.w = w;
	info.h = h;
	targetCalcCoords();

	return;
}


void renderAdjustRotation(int group, int quad) {
	float rot_x, rot_y, rot_z;

	info.group[group].qc->coord[quad].rot_x += info.group[group].quad[quad].rotvel_x;
	info.group[group].qc->coord[quad].rot_y += info.group[group].quad[quad].rotvel_y;
	info.group[group].qc->coord[quad].rot_z += info.group[group].quad[quad].rotvel_z;

	rot_x = fmodf(info.group[group].qc->coord[quad].rot_x, M_PI * 2);
	rot_y = fmodf(info.group[group].qc->coord[quad].rot_y, M_PI * 2);
	rot_z = fmodf(info.group[group].qc->coord[quad].rot_z, M_PI * 2);

	renderCacheRotateOne(info.group[group].qc, quad, rot_x, rot_y, rot_z);

	return;
}


int render() {
	int i, j;

	for (i = 0; i < info.groups; i++) {
		for (j = 0; j < info.group[i].quads; j++) {
			if (info.assemble.quad[info.group[i].quad[j].index].assembled) {
				textureUpdateNoiseProfile(&info.group[i].quad[j].noise, info.group[i].type);
			}
			genTexture(info.group[i].texture, j, &info.group[i].quad[j].noise);
			renderAdjustRotation(i, j);
		}

		renderCache(info.group[i].qc);
	}

	return 0;
}
	

void renderSpinAround(int i) {
	float x, y, z, angle;
	angle = M_PI / 500.f * i;
 
	x = ROTATE_RADIUS * cosf(angle);
	z = ROTATE_RADIUS * sinf(angle);
	x += TILE_CENTER_X;
	z += TILE_CENTER_Z;
	y = TILE_CENTER_Y;

	glLoadIdentity();
	gluPerspective(45.0f, (float) VIDEO_RES_W / (float) VIDEO_RES_H, 0.1f, 30.0f);
	gluLookAt(x, y, z, TILE_CENTER_X, TILE_CENTER_Y, TILE_CENTER_Z, 0.0f, 1.0f, 0.0f);

	return;
}


void fadeDo(int invert) {
	float coords[] = { -1.0f, 1.0f, 1.0f, 1.0f, 1.0f, -1.0f,
				1.0f, -1.0f, -1.0f, -1.0f, -1.0f, 1.0f };
	float alpha;
	glLoadIdentity();
	alpha = 1.0f / FADE_TIME * (SDL_GetTicks() - info.fade_start);
	if (invert)
		alpha = 1.0f - alpha;
	glColor4f(0.0f, 0.0f, 0.0f, alpha);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisable(GL_TEXTURE_2D);

	glVertexPointer(2, GL_FLOAT, 20, &coords);
	glDrawArrays(GL_TRIANGLES, 0, 6);

	glEnable(GL_TEXTURE_2D);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	return;
}


void timerStart() {
	info.timer = 0;
}


void fadeIn() {
	fadeDo(0);
}


void fadeOut() {
	fprintf(stderr, "Fading...\n");
	fadeDo(1);
}


void state(int i) {
/*	if (info.mode != info.newmode) {
		if (info.fade_start == 0) {
			info.fade_start = SDL_GetTicks();
		} else if (SDL_GetTicks() - info.fade_start >= FADE_TIME) {
			fadeIn();
			info.mode = info.newmode;
			info.fade_start = SDL_GetTicks();
		} else {
			fadeIn();
		}
	} else if (SDL_GetTicks() - info.fade_start < FADE_TIME)
		fadeOut();
	else
		info.fade_start = 0;
*/
	if (info.mode != info.newmode) {
		timerStart();
		info.mode = info.newmode;
	}

	switch (info.mode) {
		case MODE_FOLLOW:
			followQuad();
			if (info.timer >= FOLLOW_TIME)
				info.newmode = MODE_SPIN_AROUND;
			break;
		case MODE_SPIN_AROUND:
			renderSpinAround(i);
			if (info.timer >= SPIN_AROUND_TIME)
				info.newmode = MODE_ASSEMBLE;
			break;
		case MODE_ASSEMBLE:
			assemble();
			break;
		default:
			break;
	}

	return;
}


void timeTrackInit() {
	info.time.last = SDL_GetTicks();
	info.time.this = SDL_GetTicks();
	info.time.time = 0;

	return;
}


void timeTrack() {
	info.time.last = info.time.this;
	info.time.this = SDL_GetTicks();
	info.time.time = info.time.this - info.time.last;
	info.timer += info.time.time;

	return;
}


int main(int argc, char **argv) {
	int i;
	videoInit();

	srand(time(NULL));
	targetPatternLoad(argv[0]);
	timeTrackInit();
//	gluPerspective(45.0f, (float) VIDEO_RES_W / (float) VIDEO_RES_H, 0.1f, 30.0f);
//	gluLookAt(0.0f, 0.0f, -5.0f, TILE_CENTER_X, TILE_CENTER_Y, TILE_CENTER_Z, 0.0f, 1.0f, 0.0f);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);


	for (i = 0;; i++) {
//		renderSpinAround(i);
//		assemble();
//		followQuad();
		timeTrack();
		processEvents();
		state(i);
		render();
		SDL_GL_SwapBuffers();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		if (info.time.time < 20)
			SDL_Delay(20 - info.time.time);
	}
	/* TODO: Fortsätt */

	return 0;
}
