#include "sara.h"


int textureNoiseProfileSetup(NOISE_PROFILE *ns) {
	ns->add_r = 0;
	ns->add_g = 0;
	ns->add_b = 0;
	ns->add_a = 255;

	ns->amp_r = 255;
	ns->amp_g = 255;
	ns->amp_b = 255;
	ns->amp_a = 1;

	return 0;
}


void textureUpdateNoiseProfile(NOISE_PROFILE *ns, unsigned int target) {
	int r, g, b, a;

	r = target & 0xFF;
	g = (target >> 8) & 0xFF;
	b = (target >> 16) & 0xFF;
	a = (target >> 24) & 0xFF;

	if (ns->amp_r > 1)
		ns->amp_r--;
	if (ns->amp_g > 1)
		ns->amp_g--;
	if (ns->amp_b > 1)
		ns->amp_b--;
	if (ns->amp_a > 1)
		ns->amp_a--;
	
	if (ns->add_r - r != 0)
		ns->add_r += (ns->add_r - r < 0) ? 1 : -1;
	if (ns->add_g - g != 0)
		ns->add_g += (ns->add_g - g < 0) ? 1 : -1;
	if (ns->add_b - b != 0)
		ns->add_b += (ns->add_b - b < 0) ? 1 : -1;
	if (ns->add_a - a != 0)
		ns->add_a += (ns->add_a - a < 0) ? 1 : -1;
	return;
}



int genTexture(RENDER_TEXTURE *texture, int tile, NOISE_PROFILE *ns) {
	int x, y, w, h, i;
	int r, g, b, a;
	unsigned int *data;

	w = texture->w / texture->tile_w;
	h = texture->h / texture->tile_h;
	x = (tile % texture->tile_w) * (texture->w / texture->tile_w);
	y = (tile / texture->tile_h) * (texture->h / texture->tile_h);

	data = malloc(sizeof(unsigned int) * w * h);
	for (i = 0; i < w*h; i++) {
		r = rand() % ns->amp_r + ns->add_r;
		g = rand() % ns->amp_g + ns->add_g;
		b = rand() % ns->amp_b + ns->add_b;
		a = rand() % ns->amp_a + ns->add_a;
		data[i] = (a & 0xFF) << 24 | (b & 0xFF) << 16 | (g & 0xFF) << 8 | (r & 0xFF);
	}
	renderTextureUpdate(texture, x, y, w, h, data);
	free(data);

	return 0;
}
